#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import typing
from dataclasses import dataclass
from typing import NoReturn, Optional, TypedDict

import requests
import requests.auth

from pysonar_scanner.configuration.properties import (
    SONAR_HOST_URL,
    SONAR_SCANNER_SONARCLOUD_URL,
    SONAR_SCANNER_API_BASE_URL,
    SONAR_REGION,
    Key,
)
from pysonar_scanner.utils import remove_trailing_slash, OsStr, ArchStr
from pysonar_scanner.exceptions import (
    SonarQubeApiException,
    InconsistentConfiguration,
    SonarQubeApiUnauthroizedException,
)

GLOBAL_SONARCLOUD_URL = "https://sonarcloud.io"
US_SONARCLOUD_URL = "https://sonarqube.us"

UNAUTHORIZED_STATUS_CODES = (401, 403)

ACCEPT_JSON = {"Accept": "application/json"}
ACCEPT_OCTET_STREAM = {"Accept": "application/octet-stream"}


@dataclass(frozen=True)
class SQVersion:
    parts: list[str]

    def __get_part(self, index: int) -> int:
        if index >= len(self.parts):
            return 0
        part = self.parts[index]
        if not part.isdigit():
            return 0
        return int(part)

    def major(self) -> int:
        return self.__get_part(0)

    def minor(self) -> int:
        return self.__get_part(1)

    def does_support_bootstrapping(self) -> bool:
        if len(self.parts) == 0:
            return False

        return self.major() > MIN_SUPPORTED_SQ_VERSION.major() or (
            self.major() == MIN_SUPPORTED_SQ_VERSION.major() and self.minor() >= MIN_SUPPORTED_SQ_VERSION.minor()
        )

    def __str__(self) -> str:
        return ".".join(self.parts)

    @staticmethod
    def from_str(version: str) -> "SQVersion":
        return SQVersion(version.split("."))


MIN_SUPPORTED_SQ_VERSION: SQVersion = SQVersion.from_str("10.6")


@dataclass(frozen=True)
class BaseUrls:
    base_url: str
    api_base_url: str
    is_sonar_qube_cloud: bool

    def __post_init__(self):
        object.__setattr__(self, "base_url", remove_trailing_slash(self.base_url))
        object.__setattr__(self, "api_base_url", remove_trailing_slash(self.api_base_url))


@dataclass(frozen=True)
class JRE:
    id: Optional[str]
    filename: str
    sha256: str
    java_path: str
    os: str
    arch: str
    download_url: Optional[str]

    @staticmethod
    def from_dict(dict: dict) -> "JRE":
        return JRE(
            id=dict.get("id", None),
            filename=dict["filename"],
            sha256=dict["sha256"],
            java_path=dict["javaPath"],
            os=dict["os"],
            arch=dict["arch"],
            download_url=dict.get("downloadUrl", None),
        )


ApiConfiguration = TypedDict(
    "ApiConfiguration",
    {SONAR_HOST_URL: str, SONAR_SCANNER_SONARCLOUD_URL: str, SONAR_SCANNER_API_BASE_URL: str, SONAR_REGION: str},
)


def to_api_configuration(config_dict: dict[Key, any]) -> ApiConfiguration:
    return {
        SONAR_HOST_URL: config_dict.get(SONAR_HOST_URL, ""),
        SONAR_SCANNER_SONARCLOUD_URL: config_dict.get(SONAR_SCANNER_SONARCLOUD_URL, ""),
        SONAR_SCANNER_API_BASE_URL: config_dict.get(SONAR_SCANNER_API_BASE_URL, ""),
        SONAR_REGION: config_dict.get(SONAR_REGION, ""),
    }


def get_base_urls(config_dict: dict[Key, any]) -> BaseUrls:
    def is_sq_cloud_url(api_config: ApiConfiguration, sonar_host_url: str) -> bool:
        sq_cloud_url = api_config[SONAR_SCANNER_SONARCLOUD_URL] or GLOBAL_SONARCLOUD_URL
        return remove_trailing_slash(sonar_host_url) in [remove_trailing_slash(sq_cloud_url), US_SONARCLOUD_URL]

    def is_blank(str) -> bool:
        return str.strip() == ""

    api_config: ApiConfiguration = to_api_configuration(config_dict)

    sonar_host_url = remove_trailing_slash(api_config[SONAR_HOST_URL])
    region = api_config[SONAR_REGION]
    if region and region != "us":
        raise InconsistentConfiguration(
            f"Invalid region '{region}'. Valid regions are: 'us'. "
            "Please check the 'sonar.region' property or the 'SONAR_REGION' environment variable."
        )
    if is_inconsistent_configuration(region, sonar_host_url):
        raise InconsistentConfiguration(
            "Inconsistent values for properties 'sonar.region' and 'sonar.host.url'. "
            + "Please only specify one of the two properties."
        )
    if is_blank(sonar_host_url) or is_sq_cloud_url(api_config, sonar_host_url):
        default_url = US_SONARCLOUD_URL if region == "us" else GLOBAL_SONARCLOUD_URL
        default_api_base_url = "https://api.sonarqube.us" if region == "us" else "https://api.sonarcloud.io"
        sonar_host_url = api_config[SONAR_SCANNER_SONARCLOUD_URL] or default_url
        api_base_url = api_config[SONAR_SCANNER_API_BASE_URL] or default_api_base_url
        return BaseUrls(base_url=sonar_host_url, api_base_url=api_base_url, is_sonar_qube_cloud=True)
    else:
        api_base_url = api_config[SONAR_SCANNER_API_BASE_URL] or f"{sonar_host_url}/api/v2"
        return BaseUrls(base_url=sonar_host_url, api_base_url=api_base_url, is_sonar_qube_cloud=False)


def is_inconsistent_configuration(region, sonar_host_url):
    if not region or not sonar_host_url:
        return False
    if region == "us" and sonar_host_url.startswith(US_SONARCLOUD_URL):
        return False
    return True


class BearerAuth(requests.auth.AuthBase):
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers["Authorization"] = f"Bearer {self.token}"
        return r


@dataclass(frozen=True)
class EngineInfo:
    filename: str
    sha256: str
    download_url: Optional[str] = None


class SonarQubeApi:
    def __init__(self, base_urls: BaseUrls, token: str):
        self.base_urls = base_urls
        self.auth = BearerAuth(token)

    def __raise_exception(self, exception: Exception) -> NoReturn:
        if (
            isinstance(exception, requests.RequestException)
            and exception.response is not None
            and exception.response.status_code in UNAUTHORIZED_STATUS_CODES
        ):
            raise SonarQubeApiUnauthroizedException.create_default(self.base_urls.base_url) from exception
        else:
            raise SonarQubeApiException("Error while fetching the analysis version") from exception

    def is_sonar_qube_cloud(self) -> bool:
        return self.base_urls.is_sonar_qube_cloud

    def get_analysis_version(self) -> SQVersion:
        try:
            res = requests.get(f"{self.base_urls.api_base_url}/analysis/version", auth=self.auth)
            if res.status_code != 200:
                res = requests.get(f"{self.base_urls.base_url}/api/server/version", auth=self.auth)

            res.raise_for_status()
            return SQVersion.from_str(res.text)
        except requests.RequestException as e:
            self.__raise_exception(e)

    def get_analysis_engine(self) -> EngineInfo:
        try:
            res = requests.get(f"{self.base_urls.api_base_url}/analysis/engine", headers=ACCEPT_JSON, auth=self.auth)
            res.raise_for_status()
            json = res.json()
            if "filename" not in json or "sha256" not in json:
                raise SonarQubeApiException("Invalid response from the server")
            return EngineInfo(
                filename=json["filename"], sha256=json["sha256"], download_url=json.get("downloadUrl", None)
            )
        except requests.RequestException as e:
            self.__raise_exception(e)

    def download_analysis_engine(self, handle: typing.BinaryIO) -> None:
        """
        This method can raise a SonarQubeApiException if the server doesn't respond successfully.
        Alternative, if the file IO fails, an IOError or OSError can be raised.
        """
        try:
            res = requests.get(
                f"{self.base_urls.api_base_url}/analysis/engine",
                headers=ACCEPT_OCTET_STREAM,
                auth=self.auth,
            )
            self.__download_file(res, handle)
        except requests.RequestException as e:
            self.__raise_exception(e)

    def get_analysis_jres(self, os: OsStr, arch: ArchStr) -> list[JRE]:
        try:
            params = {"os": os, "arch": arch}
            res = requests.get(
                f"{self.base_urls.api_base_url}/analysis/jres",
                auth=self.auth,
                headers=ACCEPT_JSON,
                params=params,
            )
            res.raise_for_status()
            json_array = res.json()
            return [JRE.from_dict(jre) for jre in json_array]
        except (requests.RequestException, KeyError) as e:
            self.__raise_exception(e)

    def download_analysis_jre(self, id: str, handle: typing.BinaryIO) -> None:
        """
        This method can raise a SonarQubeApiException if the server doesn't respond successfully.
        Alternative, if the file IO fails, an IOError or OSError can be raised.
        """

        try:
            res = requests.get(
                f"{self.base_urls.api_base_url}/analysis/jres/{id}",
                headers=ACCEPT_OCTET_STREAM,
                auth=self.auth,
            )
            self.__download_file(res, handle)
        except requests.RequestException as e:
            self.__raise_exception(e)

    def download_file_from_url(self, url: str, handle: typing.BinaryIO) -> None:
        """
        This method can raise a SonarQubeApiException if the server doesn't respond successfully.
        Alternative, if the file IO fails, an IOError or OSError can be raised.
        """
        try:
            res = requests.get(
                url,
                headers=ACCEPT_OCTET_STREAM,
            )
            self.__download_file(res, handle)
        except requests.RequestException as e:
            self.__raise_exception(e)

    def __download_file(self, res: requests.Response, handle: typing.BinaryIO) -> None:
        res.raise_for_status()
        for chunk in res.iter_content(chunk_size=128):
            handle.write(chunk)



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import hashlib
import pathlib
import platform
import typing
from enum import Enum

OsStr = typing.Literal["windows", "linux", "mac", "alpine", "other"]
ArchStr = typing.Literal["x64", "aarch64", "other"]


def remove_trailing_slash(url: str) -> str:
    return url.rstrip("/ ").lstrip()


def calculate_checksum(filehandle: typing.BinaryIO) -> str:
    sha256_hash = hashlib.sha256()
    for byte_block in iter(lambda: filehandle.read(4096), b""):
        sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


class Os(Enum):
    WINDOWS: OsStr = "windows"
    LINUX: OsStr = "linux"
    MACOS: OsStr = "mac"
    ALPINE: OsStr = "alpine"
    OTHER: OsStr = "other"


def get_os() -> Os:
    def is_alpine() -> bool:
        try:
            os_release = pathlib.Path("/etc/os-release")
            if not os_release.exists():
                os_release = pathlib.Path("/usr/lib/os-release")
            if not os_release.exists():
                return False
            os_release_str = os_release.read_text()
            return "ID=alpine" in os_release_str or 'ID="alpine"' in os_release_str
        except OSError:
            return False

    os_name = platform.system()
    if os_name == "Windows":
        return Os.WINDOWS
    elif os_name == "Darwin":
        return Os.MACOS
    elif os_name == "Linux":
        if is_alpine():
            return Os.ALPINE
        else:
            return Os.LINUX
    return Os.OTHER


class Arch(Enum):
    X64: ArchStr = "x64"
    AARCH64: ArchStr = "aarch64"
    OTHER: ArchStr = "other"


def get_arch() -> Arch:
    machine = platform.machine().lower()
    if machine in ["amd64", "x86_64"]:
        return Arch.X64
    elif machine == "arm64":
        return Arch.AARCH64
    return Arch.OTHER


def filter_none_values(dictionary: dict) -> dict:
    return {k: v for k, v in dictionary.items() if v is not None}



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import os
from typing import Dict

from pysonar_scanner.configuration.properties import Key, SONAR_SCANNER_OS, SONAR_SCANNER_ARCH, SONAR_PROJECT_BASE_DIR
from pysonar_scanner import utils


def load() -> Dict[Key, any]:
    """
    Load dynamically computed default properties
    """
    properties = {
        SONAR_SCANNER_OS: utils.get_os().value,
        SONAR_SCANNER_ARCH: utils.get_arch().value,
        SONAR_PROJECT_BASE_DIR: os.getcwd(),
    }
    return properties



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import logging
from pysonar_scanner import app_logging
from pysonar_scanner import cache
from pysonar_scanner import exceptions
from pysonar_scanner.api import get_base_urls, SonarQubeApi, BaseUrls, MIN_SUPPORTED_SQ_VERSION
from pysonar_scanner.configuration import configuration_loader
from pysonar_scanner.configuration.configuration_loader import ConfigurationLoader
from pysonar_scanner.configuration.properties import (
    SONAR_VERBOSE,
    SONAR_HOST_URL,
    SONAR_SCANNER_API_BASE_URL,
    SONAR_SCANNER_SONARCLOUD_URL,
    SONAR_SCANNER_PROXY_PORT,
    SONAR_SCANNER_JAVA_EXE_PATH,
    SONAR_SCANNER_OS,
    SONAR_SCANNER_ARCH,
)
from pysonar_scanner.exceptions import SQTooOldException
from pysonar_scanner.jre import JREResolvedPath, JREProvisioner, JREResolver, JREResolverConfiguration
from pysonar_scanner.scannerengine import ScannerEngine, ScannerEngineProvisioner


def scan():
    try:
        return do_scan()
    except Exception as e:
        return exceptions.log_error(e)


def do_scan():
    app_logging.setup()
    logging.info("Starting Pysonar, the Sonar scanner CLI for Python")
    config = ConfigurationLoader.load()
    set_logging_options(config)

    ConfigurationLoader.check_configuration(config)

    api = build_api(config)
    check_version(api)
    update_config_with_api_urls(config, api.base_urls)
    logging.debug(f"Final loaded configuration: {config}")

    cache_manager = cache.get_cache(config)
    scanner = create_scanner_engine(api, cache_manager, config)

    logging.info("Starting the analysis...")
    return scanner.run(config)


def set_logging_options(config):
    app_logging.configure_logging_level(verbose=config.get(SONAR_VERBOSE, False))


def build_api(config: dict[str, any]) -> SonarQubeApi:
    token = configuration_loader.get_token(config)
    base_urls = get_base_urls(config)
    return SonarQubeApi(base_urls, token)


def check_version(api: SonarQubeApi):
    if api.is_sonar_qube_cloud():
        logging.debug(f"SonarQube Cloud url: {api.base_urls.base_url}")
        return
    version = api.get_analysis_version()
    logging.debug(f"SonarQube url: {api.base_urls.base_url}")

    if not version.does_support_bootstrapping():
        raise SQTooOldException(
            f"This scanner only supports SonarQube versions >= {MIN_SUPPORTED_SQ_VERSION}. \n"
            f"The server at {api.base_urls.base_url} is on version {version}\n"
            "Please either upgrade your SonarQube server or use the Sonar Scanner CLI (see https://docs.sonarsource.com/sonarqube-server/latest/analyzing-source-code/scanners/sonarscanner/)."
        )


def update_config_with_api_urls(config, base_urls: BaseUrls):
    config[SONAR_HOST_URL] = base_urls.base_url
    config[SONAR_SCANNER_API_BASE_URL] = base_urls.api_base_url
    if base_urls.is_sonar_qube_cloud:
        config[SONAR_SCANNER_SONARCLOUD_URL] = base_urls.base_url
    config[SONAR_SCANNER_PROXY_PORT] = "443" if base_urls.base_url.startswith("https") else "80"


def create_scanner_engine(api, cache_manager, config):
    jre_path = create_jre(api, cache_manager, config)
    config[SONAR_SCANNER_JAVA_EXE_PATH] = str(jre_path.path)
    logging.debug(f"JRE path: {jre_path.path}")
    scanner_engine_path = ScannerEngineProvisioner(api, cache_manager).provision()
    scanner = ScannerEngine(jre_path, scanner_engine_path)
    return scanner


def create_jre(api, cache, config: dict[str, any]) -> JREResolvedPath:
    jre_provisioner = JREProvisioner(api, cache, config[SONAR_SCANNER_OS], config[SONAR_SCANNER_ARCH])
    jre_resolver = JREResolver(JREResolverConfiguration.from_dict(config), jre_provisioner)
    return jre_resolver.resolve_jre()



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import time
import argparse
from dataclasses import dataclass
from typing import Callable, Optional

Key = str

SONAR_HOST_URL: Key = "sonar.host.url"
SONAR_SCANNER_SONARCLOUD_URL: Key = "sonar.scanner.sonarcloudUrl"
SONAR_SCANNER_API_BASE_URL: Key = "sonar.scanner.apiBaseUrl"
SONAR_REGION: Key = "sonar.region"
SONAR_ORGANIZATION: Key = "sonar.organization"
SONAR_SCANNER_APP: Key = "sonar.scanner.app"
SONAR_SCANNER_APP_VERSION: Key = "sonar.scanner.appVersion"
SONAR_SCANNER_BOOTSTRAP_START_TIME: Key = "sonar.scanner.bootstrapStartTime"
SONAR_SCANNER_WAS_JRE_CACHE_HIT: Key = "sonar.scanner.wasJreCacheHit"
SONAR_SCANNER_WAS_ENGINE_CACHE_HIT: Key = "sonar.scanner.wasEngineCacheHit"
SONAR_VERBOSE: Key = "sonar.verbose"
SONAR_TOKEN: Key = "sonar.token"
SONAR_SCANNER_OS: Key = "sonar.scanner.os"
SONAR_SCANNER_ARCH: Key = "sonar.scanner.arch"
SONAR_SCANNER_SKIP_JRE_PROVISIONING: Key = "sonar.scanner.skipJreProvisioning"
SONAR_USER_HOME: Key = "sonar.userHome"
SONAR_SCANNER_JAVA_EXE_PATH: Key = "sonar.scanner.javaExePath"
SONAR_SCANNER_INTERNAL_DUMP_TO_FILE: Key = "sonar.scanner.internal.dumpToFile"
SONAR_SCANNER_INTERNAL_SQ_VERSION: Key = "sonar.scanner.internal.sqVersion"
SONAR_SCANNER_CONNECT_TIMEOUT: Key = "sonar.scanner.connectTimeout"
SONAR_SCANNER_SOCKET_TIMEOUT: Key = "sonar.scanner.socketTimeout"
SONAR_SCANNER_RESPONSE_TIMEOUT: Key = "sonar.scanner.responseTimeout"
SONAR_SCANNER_TRUSTSTORE_PATH: Key = "sonar.scanner.truststorePath"
SONAR_SCANNER_TRUSTSTORE_PASSWORD: Key = "sonar.scanner.truststorePassword"
SONAR_SCANNER_KEYSTORE_PATH: Key = "sonar.scanner.keystorePath"
SONAR_SCANNER_KEYSTORE_PASSWORD: Key = "sonar.scanner.keystorePassword"
SONAR_SCANNER_PROXY_HOST: Key = "sonar.scanner.proxyHost"
SONAR_SCANNER_PROXY_PORT: Key = "sonar.scanner.proxyPort"
SONAR_SCANNER_PROXY_USER: Key = "sonar.scanner.proxyUser"
SONAR_SCANNER_PROXY_PASSWORD: Key = "sonar.scanner.proxyPassword"
SONAR_SCANNER_METADATA_FILEPATH: Key = "sonar.scanner.metadataFilePath"
SONAR_SCANNER_JAVA_OPTS: Key = "sonar.scanner.javaOpts"
SONAR_PROJECT_BASE_DIR: Key = "sonar.projectBaseDir"
SONAR_PROJECT_KEY: Key = "sonar.projectKey"
SONAR_PROJECT_NAME: Key = "sonar.projectName"
SONAR_PROJECT_VERSION: Key = "sonar.projectVersion"
SONAR_PROJECT_DESCRIPTION: Key = "sonar.projectDescription"
SONAR_SCM_EXCLUSIONS_DISABLED = "sonar.scm.exclusions.disabled"
SONAR_SOURCES: Key = "sonar.sources"
SONAR_EXCLUSIONS: Key = "sonar.exclusions"
SONAR_TESTS: Key = "sonar.tests"
SONAR_FILESIZE_LIMIT: Key = "sonar.filesize.limit"
SONAR_CPD_PYTHON_MINIMUM_TOKENS: Key = "sonar.cpd.python.minimumTokens"
SONAR_CPD_PYTHON_MINIMUM_LINES: Key = "sonar.cpd.python.minimumLines"
SONAR_LOG_LEVEL: Key = "sonar.log.level"
SONAR_QUALITYGATE_WAIT: Key = "sonar.qualitygate.wait"
SONAR_QUALITYGATE_TIMEOUT: Key = "sonar.qualitygate.timeout"
SONAR_EXTERNAL_ISSUES_REPORT_PATHS: Key = "sonar.externalIssuesReportPaths"
SONAR_SARIF_REPORT_PATHS: Key = "sonar.sarifReportPaths"
SONAR_LINKS_CI: Key = "sonar.links.ci"
SONAR_LINKS_HOMEPAGE: Key = "sonar.links.homepage"
SONAR_LINKS_ISSUE: Key = "sonar.links.issue"
SONAR_LINKS_SCM: Key = "sonar.links.scm"
SONAR_BRANCH_NAME: Key = "sonar.branch.name"
SONAR_PULLREQUEST_KEY: Key = "sonar.pullrequest.key"
SONAR_PULLREQUEST_BRANCH: Key = "sonar.pullrequest.branch"
SONAR_PULLREQUEST_BASE: Key = "sonar.pullrequest.base"
SONAR_PYTHON_VERSION: Key = "sonar.python.version"
SONAR_PYTHON_PYLINT_REPORT_PATH: Key = "sonar.python.pylint.reportPath"
SONAR_PYTHON_COVERAGE_REPORT_PATHS: Key = "sonar.python.coverage.reportPaths"
SONAR_PYTHON_SKIP_UNCHANGED: Key = "sonar.python.skipUnchanged"
SONAR_NEWCODE_REFERENCE_BRANCH: Key = "sonar.newCode.referenceBranch"
SONAR_SCM_REVISION: Key = "sonar.scm.revision"
SONAR_BUILD_STRING: Key = "sonar.buildString"
SONAR_SOURCE_ENCODING: Key = "sonar.sourceEncoding"
SONAR_WORKING_DIRECTORY: Key = "sonar.working.directory"
SONAR_SCM_FORCE_RELOAD_ALL: Key = "sonar.scm.forceReloadAll"
SONAR_MODULES: Key = "sonar.modules"
SONAR_PYTHON_XUNIT_REPORT_PATH = "sonar.python.xunit.reportPath"
SONAR_PYTHON_XUNIT_SKIP_DETAILS = "sonar.python.xunit.skipDetails"
SONAR_PYTHON_MYPY_REPORT_PATHS = "sonar.python.mypy.reportPaths"
SONAR_PYTHON_BANDIT_REPORT_PATHS = "sonar.python.bandit.reportPaths"
SONAR_PYTHON_FLAKE8_REPORT_PATHS = "sonar.python.flake8.reportPaths"
SONAR_PYTHON_RUFF_REPORT_PATHS = "sonar.python.ruff.reportPaths"
TOML_PATH: Key = "toml-path"


@dataclass
class Property:
    name: Key
    """name in the format of `sonar.scanner.appVersion`"""

    default_value: Optional[any]
    """default value for the property; if None, no default value is set"""

    cli_getter: Optional[Callable[[argparse.Namespace], any]] = None
    """function to get the value from the CLI arguments namespace. If None, the property is not settable via CLI"""

    def python_name(self) -> str:
        """Convert Java-style camel case name to Python-style kebab-case name."""
        result = []
        for i, char in enumerate(self.name):
            if char.isupper() and i > 0:
                result.append("-")
            result.append(char.lower())
        return "".join(result)

    def env_variable_name(self) -> str:
        """Convert property name to environment variable name format.
        Example: sonar.scanner.proxyPort -> SONAR_SCANNER_PROXY_PORT"""
        # Replace dots with underscores
        env_name = self.name.replace(".", "_")

        # Insert underscores before uppercase letters (camelCase to snake_case)
        result = []
        for i, char in enumerate(env_name):
            if char.isupper() and i > 0 and env_name[i - 1] != "_":
                result.append("_")
            result.append(char)
        # Convert to uppercase
        return "".join(result).upper()


# fmt: off
PROPERTIES: list[Property] = [
    Property(
        name=SONAR_SCANNER_APP, 
        default_value="python", 
        cli_getter=None
    ),
    Property(
        name=SONAR_SCANNER_APP_VERSION, 
        default_value="1.0", 
        cli_getter=None
    ),
    Property(
        name=SONAR_SCANNER_BOOTSTRAP_START_TIME,
        default_value=int(time.time() * 1000),
        cli_getter=None
    ),
    Property(
        name=SONAR_SCANNER_WAS_JRE_CACHE_HIT, 
        default_value=None, 
        cli_getter=None
    ),
    Property(
        name=SONAR_SCANNER_WAS_ENGINE_CACHE_HIT, 
        default_value=None, 
        cli_getter=None
    ),
    Property(
        name=SONAR_VERBOSE, 
        default_value=False, 
        cli_getter=lambda args: args.verbose
    ),
    Property(
        name=SONAR_HOST_URL, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_host_url
    ),  
    Property(
        name=SONAR_REGION, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_region
    ),
    Property(
        name=SONAR_ORGANIZATION, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_organization
    ),
    Property(
        name=SONAR_SCANNER_SONARCLOUD_URL,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_cloud_url
    ),  
    Property(
        name=SONAR_SCANNER_API_BASE_URL, 
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_api_url
    ),  
    Property(
        name=SONAR_TOKEN, 
        default_value=None, 
        cli_getter=lambda args: args.token
    ),
    Property(
        name=SONAR_SCANNER_OS, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_os
    ),
    Property(
        name=SONAR_SCANNER_ARCH, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_arch
    ),
    Property(
        name=SONAR_SCANNER_SKIP_JRE_PROVISIONING,
        default_value=False,
        cli_getter=lambda args: args.skip_jre_provisioning
    ),
    Property(
        name=SONAR_USER_HOME, 
        default_value=None,
        cli_getter=lambda args: args.sonar_user_home
    ),  
    Property(
        name=SONAR_SCANNER_JAVA_EXE_PATH, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_java_exe_path
    ),
    Property(
        name=SONAR_SCANNER_INTERNAL_DUMP_TO_FILE,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_internal_dump_to_file
    ),
    Property(
        name=SONAR_SCANNER_INTERNAL_SQ_VERSION,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_internal_sq_version
    ),
    Property(
        name=SONAR_SCANNER_CONNECT_TIMEOUT,
        default_value=5,
        cli_getter=lambda args: args.sonar_scanner_connect_timeout
    ),
    Property(
        name=SONAR_SCANNER_SOCKET_TIMEOUT,
        default_value=60,
        cli_getter=lambda args: args.sonar_scanner_socket_timeout
    ),
    Property(
        name=SONAR_SCANNER_RESPONSE_TIMEOUT,
        default_value=0,
        cli_getter=lambda args: args.sonar_scanner_response_timeout
    ),
    Property(
        name=SONAR_SCANNER_TRUSTSTORE_PATH,
        default_value=None,  
        cli_getter=lambda args: args.sonar_scanner_truststore_path
    ),
    Property(
        name=SONAR_SCANNER_TRUSTSTORE_PASSWORD,
        default_value="changeit",  
        cli_getter=lambda args: args.sonar_scanner_truststore_password
    ),
    Property(
        name=SONAR_SCANNER_KEYSTORE_PATH,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_keystore_path
    ),  
    Property(
        name=SONAR_SCANNER_KEYSTORE_PASSWORD,
        default_value="changeit",  
        cli_getter=lambda args: args.sonar_scanner_keystore_password
    ),
    Property(
        name=SONAR_SCANNER_PROXY_HOST, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_proxy_host
    ),
    Property(
        name=SONAR_SCANNER_PROXY_PORT, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_proxy_port
    ),  
    Property(
        name=SONAR_SCANNER_PROXY_USER, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_proxy_user
    ),
    Property(
        name=SONAR_SCANNER_PROXY_PASSWORD,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_proxy_password
    ),
    Property(
        name=SONAR_PROJECT_BASE_DIR, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_project_base_dir
    ),  
    Property(
        name=SONAR_SCANNER_JAVA_OPTS, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scanner_java_opts
    ),
    Property(
        name=SONAR_SCANNER_METADATA_FILEPATH,
        default_value=None,
        cli_getter=lambda args: args.sonar_scanner_metadata_filepath
    ),
    Property(
        name=SONAR_PROJECT_KEY, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_project_key
    ),
    Property(
        name=SONAR_PROJECT_NAME,
        default_value=None,
        cli_getter=lambda args: args.sonar_project_name
    ),
    Property(
        name=SONAR_PROJECT_VERSION,
        default_value=None,
        cli_getter=lambda args: args.sonar_project_version
    ),
    Property(
        name=SONAR_PROJECT_DESCRIPTION,
        default_value=None,
        cli_getter=lambda args: args.sonar_project_description
    ),
    Property(
        name=SONAR_SOURCES,
        default_value=None,
        cli_getter=lambda args: args.sonar_sources
    ),
    Property(
        name=SONAR_TESTS,
        default_value=None,
        cli_getter=lambda args: args.sonar_tests
    ),
    Property(
        name=SONAR_FILESIZE_LIMIT,
        default_value=None,
        cli_getter=lambda args: args.sonar_filesize_limit
    ),
    Property(
        name=SONAR_SCM_EXCLUSIONS_DISABLED,
        default_value=None,
        cli_getter=lambda args: args.sonar_scm_exclusions_disabled or getattr(args, "Dsonar.scm.exclusions.disabled")
    ),
    Property(
        name=SONAR_CPD_PYTHON_MINIMUM_TOKENS,
        default_value=None,
        cli_getter=lambda args: args.sonar_cpd_python_minimum_tokens
    ),
    Property(
        name=SONAR_CPD_PYTHON_MINIMUM_LINES,
        default_value=None,
        cli_getter=lambda args: args.sonar_cpd_python_minimum_lines
    ),
    Property(
        name=SONAR_LOG_LEVEL,
        default_value=None,
        cli_getter=lambda args: args.sonar_log_level
    ),
    Property(
        name=TOML_PATH,
        default_value=None,
        cli_getter=lambda args: args.toml_path
    ),
    Property(
        name=SONAR_PROJECT_VERSION,
        default_value=None,
        cli_getter=lambda args: args.sonar_project_version
    ),
    Property(
        name=SONAR_PROJECT_DESCRIPTION,
        default_value=None,
        cli_getter=lambda args: args.sonar_project_description
    ),
    Property(
        name=SONAR_PYTHON_VERSION,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_version
    ),
    Property(
        name=SONAR_QUALITYGATE_WAIT, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_qualitygate_wait or getattr(args, "Dsonar.qualitygate.wait")
    ),
    Property(
        name=SONAR_QUALITYGATE_TIMEOUT, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_qualitygate_timeout
    ),
    Property(
        name=SONAR_EXTERNAL_ISSUES_REPORT_PATHS, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_external_issues_report_paths
    ),
    Property(
        name=SONAR_SARIF_REPORT_PATHS, 
        default_value=None,
        cli_getter=lambda args: args.sonar_sarif_report_paths
    ),
    Property(
        name=SONAR_LINKS_CI, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_links_ci
    ),
    Property(
        name=SONAR_LINKS_HOMEPAGE, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_links_homepage
    ),
    Property(
        name=SONAR_LINKS_ISSUE, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_links_issue
    ),
    Property(
        name=SONAR_LINKS_SCM, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_links_scm
    ),
    Property(
        name=SONAR_BRANCH_NAME, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_branch_name
    ),
    Property(
        name=SONAR_PULLREQUEST_KEY, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_pullrequest_key
    ),
    Property(
        name=SONAR_PULLREQUEST_BRANCH, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_pullrequest_branch
    ),
    Property(
        name=SONAR_PULLREQUEST_BASE, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_pullrequest_base
    ),
    Property(
        name=SONAR_NEWCODE_REFERENCE_BRANCH, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_newcode_reference_branch
    ),
    Property(
        name=SONAR_SCM_REVISION, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scm_revision
    ),
    Property(
        name=SONAR_BUILD_STRING, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_build_string
    ),
    Property(
        name=SONAR_SOURCE_ENCODING, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_source_encoding
    ),
    Property(
        name=SONAR_WORKING_DIRECTORY, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_working_directory
    ),
    Property(
        name=SONAR_SCM_FORCE_RELOAD_ALL, 
        default_value=None, 
        cli_getter=lambda args: args.sonar_scm_force_reload_all or getattr(args, "Dsonar.scm.forceReloadAll")
    ),
    Property(
        name=SONAR_PYTHON_PYLINT_REPORT_PATH,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_pylint_report_path
    ),
    Property(
        name=SONAR_PYTHON_COVERAGE_REPORT_PATHS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_coverage_report_paths
    ),
    Property(
        name=SONAR_PYTHON_SKIP_UNCHANGED,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_skip_unchanged or getattr(args, "Dsonar.python.skipUnchanged")
    ),
    Property(
        name=SONAR_PYTHON_XUNIT_REPORT_PATH,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_xunit_report_path
    ),
    Property(
        name=SONAR_PYTHON_XUNIT_SKIP_DETAILS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_xunit_skip_details or getattr(args, "Dsonar.python.xunit.skipDetails")
    ),
    Property(
        name=SONAR_PYTHON_MYPY_REPORT_PATHS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_mypy_report_paths
    ),
    Property(
        name=SONAR_PYTHON_BANDIT_REPORT_PATHS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_bandit_report_paths
    ),
    Property(
        name=SONAR_PYTHON_FLAKE8_REPORT_PATHS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_flake8_report_paths
    ),
    Property(
        name=SONAR_PYTHON_RUFF_REPORT_PATHS,
        default_value=None,
        cli_getter=lambda args: args.sonar_python_ruff_report_paths
    ),
    Property(
        name=SONAR_MODULES,
        default_value=None,
        cli_getter=lambda args: args.sonar_modules
    )
]
# fmt: on



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging


def setup() -> None:
    logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")


def configure_logging_level(verbose: bool) -> None:
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
import logging
from pathlib import Path
from typing import Dict
import os
import tomli

from pysonar_scanner.configuration import properties


class TomlProperties:
    sonar_properties: Dict[str, str]
    project_properties: Dict[str, str]

    def __init__(self, sonar_properties: Dict[str, str], project_properties: Dict[str, str]):
        self.sonar_properties = sonar_properties
        self.project_properties = project_properties


class TomlConfigurationLoader:
    @staticmethod
    def load(base_dir: Path) -> TomlProperties:
        filepath = base_dir / "pyproject.toml"
        if not os.path.isfile(filepath):
            logging.debug(f"No pyproject.toml at {filepath}")
            return TomlProperties({}, {})
        logging.debug(f"pyproject.toml loaded from {filepath}")
        try:
            with open(filepath, "rb") as f:
                toml_dict = tomli.load(f)
            # Look for configuration in the tool.sonar section
            sonar_properties = TomlConfigurationLoader.__read_sonar_properties(toml_dict)
            # Look for general project configuration
            project_properties = TomlConfigurationLoader.__read_project_properties(toml_dict)
            return TomlProperties(sonar_properties, project_properties)
        except Exception as e:
            logging.warning(
                f"There was an error reading the pyproject.toml file. No properties from the TOML file were extracted. Error: {e}"
            )
            return TomlProperties({}, {})

    @staticmethod
    def __read_sonar_properties(toml_dict) -> Dict[str, str]:
        if "tool" in toml_dict and "sonar" in toml_dict["tool"]:
            sonar_config = toml_dict["tool"]["sonar"]
            python_to_java_names = {prop.python_name(): prop.name for prop in properties.PROPERTIES}
            flattened_sonar_config = TomlConfigurationLoader.__flatten_config_dict(sonar_config, prefix="sonar.")
            return {
                python_to_java_names.get(key, TomlConfigurationLoader.__kebab_to_camel_case(key)): value
                for key, value in flattened_sonar_config.items()
            }
        return {}

    @staticmethod
    def __read_project_properties(toml_dict) -> Dict[str, str]:
        # Extract project metadata
        project_properties = {}
        if "project" in toml_dict:
            project_data = toml_dict["project"]
            # Known pyproject.toml project keys and associated Sonar property names
            property_mapping = {
                "name": properties.SONAR_PROJECT_NAME,
                "description": properties.SONAR_PROJECT_DESCRIPTION,
                "version": properties.SONAR_PROJECT_VERSION,
                "requires-python": properties.SONAR_PYTHON_VERSION,
            }

            for toml_key, sonar_property in property_mapping.items():
                if toml_key in project_data:
                    string_property = TomlConfigurationLoader.__convert_arrays_to_string(project_data[toml_key])
                    project_properties[sonar_property] = string_property
        return project_properties

    @staticmethod
    def __flatten_config_dict(config: dict[str, any], prefix: str) -> dict[str, any]:
        """Flatten nested dictionaries into dot notation keys"""
        result = {}
        for key, value in config.items():
            if isinstance(value, dict):
                result.update(TomlConfigurationLoader.__flatten_config_dict(value, f"{prefix}{key}."))
            else:
                property_name = f"{prefix}{key}"
                result[property_name] = value
        return result

    @staticmethod
    def __convert_arrays_to_string(property) -> str:
        if isinstance(property, list):
            return ",".join(str(item) for item in property)
        else:
            return property

    @staticmethod
    def __kebab_to_camel_case(key: str) -> str:
        if "-" in key:
            parts = key.split("-")
            result = parts[0] + "".join(word.capitalize() for word in parts[1:])
            logging.debug(f"Converting kebab-case property '{key}' to camelCase: '{result}'")
            return result
        return key



#
# Sonar Scanner Python
# Copyright (C) 2011-2024 SonarSource SA.
# mailto:info AT sonarsource DOT com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful,
#
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#

import pathlib
import typing
from dataclasses import dataclass

from pysonar_scanner import utils
from pysonar_scanner.configuration.properties import SONAR_USER_HOME

OpenBinaryMode = typing.Literal["wb", "xb"]


@dataclass(frozen=True)
class CacheFile:
    filepath: pathlib.Path
    checksum: str

    def exists(self) -> bool:
        return self.filepath.exists()

    def is_valid(self) -> bool:
        try:
            with open(self.filepath, "rb") as f:
                calculated_checksum = utils.calculate_checksum(f)

            return calculated_checksum == self.checksum
        except OSError:
            return False

    def open(self, mode: OpenBinaryMode) -> typing.BinaryIO:
        return open(self.filepath, mode=mode)


class Cache:
    def __init__(self, cache_folder: pathlib.Path):
        if not cache_folder.exists():
            raise FileNotFoundError(f"Cache folder {cache_folder} does not exist")
        self.cache_folder = cache_folder

    def get_file(self, filename: str, checksum: str) -> CacheFile:
        path = self.cache_folder / filename
        return CacheFile(path, checksum)

    def get_file_path(self, filename: str) -> pathlib.Path:
        return self.cache_folder / filename

    @staticmethod
    def create_cache(cache_folder: pathlib.Path):
        if not cache_folder.exists():
            cache_folder.mkdir(parents=True)
        return Cache(cache_folder)


def get_cache(config) -> Cache:
    if SONAR_USER_HOME in config:
        cache_folder = pathlib.Path(config[SONAR_USER_HOME]) / "cache"
    else:
        cache_folder = pathlib.Path.home() / ".sonar/cache"
    return Cache.create_cache(cache_folder)
